@@ -5,8 +5,8 @@
 {-# LANGUAGE DerivingVia           #-}
 {-# LANGUAGE FlexibleContexts      #-}
 {-# LANGUAGE FlexibleInstances     #-}
-{-# LANGUAGE KindSignatures        #-}
 {-# LANGUAGE MultiParamTypeClasses #-}
+{-# LANGUAGE NamedFieldPuns        #-}
 {-# LANGUAGE ScopedTypeVariables   #-}
 {-# LANGUAGE StandaloneDeriving    #-}
 {-# LANGUAGE TypeApplications      #-}
@@ -18,7 +18,9 @@ module NoThunks.Class (
     -- * Check a value for unexpected thunks
     NoThunks(..)
   , ThunkInfo(..)
+  , mkThunkInfo
   , Context
+  , Info
   , unsafeNoThunks
     -- * Helpers for defining instances
   , allNoThunks
@@ -58,6 +60,9 @@ import Data.Ratio
 import Data.Sequence (Seq)
 import Data.Set (Set)
 import Data.Time
+#if MIN_VERSION_base(4,16,0)
+import Data.Tuple (Solo (..))
+#endif
 import Data.Void (Void)
 import Data.Word
 import GHC.Stack
@@ -66,12 +71,18 @@ import GHC.Stack
 import Numeric.Natural
 #endif
 
+#if MIN_VERSION_base(4,16,0)
+import GHC.InfoProv.Compat
+#endif
+
 import qualified Control.Concurrent.MVar       as MVar
 import qualified Control.Concurrent.STM.TVar   as TVar
 import qualified Data.IntMap                   as IntMap
 import qualified Data.IORef                    as IORef
 import qualified Data.Map                      as Map
 import qualified Data.Set                      as Set
+import qualified Data.Monoid                   as Monoid
+import qualified Data.Semigroup                as Semigroup
 
 #ifdef MIN_VERSION_bytestring
 import Data.ByteString.Short (ShortByteString)
@@ -119,20 +130,20 @@ class NoThunks a where
   -- reports the /unexpected/ thunks.
   --
   -- The default implementation of 'noThunks' checks that the argument is in
-  -- WHNF, and if so, adds the type into the context (using 'showTypeOf'), and
-  -- calls 'wNoThunks'. See 'ThunkInfo' for a detailed discussion of the type
-  -- context.
+  -- WHNF, and if so, adds the type into the context (using 'showTypeOf' or
+  -- 'whereFrom' if available), and calls 'wNoThunks'. See 'ThunkInfo' for
+  -- a detailed discussion of the type context.
+  --
   --
   -- See also discussion of caveats listed for 'checkContainsThunks'.
   noThunks :: Context -> a -> IO (Maybe ThunkInfo)
   noThunks ctxt x = do
       isThunk <- checkIsThunk x
+      let ctxt' = showTypeOf (Proxy @a) : ctxt
+      thunkInfo <- mkThunkInfo ctxt' x
       if isThunk
-        then return $ Just ThunkInfo { thunkContext = ctxt' }
+        then return $ Just thunkInfo
         else wNoThunks ctxt' x
-    where
-      ctxt' :: Context
-      ctxt' = showTypeOf (Proxy @a) : ctxt
 
   -- | Check that the argument is in normal form, assuming it is in WHNF.
   --
@@ -185,32 +196,59 @@ class NoThunks a where
 -- @Int@ which was a thunk.
 type Context = [String]
 
+-- | Binding name, type and location information about the thunk, e.g.
+--
+-- > fromModel :: Int @ test/Test/NoThunks/Class.hs:198:53-84
+--
+type Info = String
+
 {-------------------------------------------------------------------------------
   Results of the check
 -------------------------------------------------------------------------------}
 
 -- | Information about unexpected thunks
 --
--- TODO: The ghc-debug work by Matthew Pickering includes some work that allows
--- to get source spans from closures. If we could take advantage of that, we
--- could not only show the type of the unexpected thunk, but also where it got
--- allocated.
-newtype ThunkInfo = ThunkInfo {
-      -- The @Context@ argument is intended to give a clue to add debugging.
-      -- For example, suppose we have something of type @(Int, [Int])@. The
-      -- various contexts we might get are
-      --
-      -- > Context                  The thunk is..
-      -- > ---------------------------------------------------------------------
-      -- > ["(,)"]                  the pair itself
-      -- > ["Int","(,)"]            the Int in the pair
-      -- > ["List","(,)"]           the [Int] in the pair
-      -- > ["Int","List","(,)"]     an Int in the [Int] in the pair
-      --
-      -- Note: prior to `ghc-9.6` a list was indicated by `[]`.
+-- ThunkInfo contains either precise `Info` about the thunk location
+-- or `Context` to make it easier to debug space leaks.  `Info` is available if
+--
+-- * @GHC-9.4@ or newer is used,
+-- * the code is compiled with @-finfo-table-map@ and is improved if
+--   @-fdistinct-constructor-tables@ is used as well.
+--
+-- The @Context@ argument is intended to give a clue to add debugging.
+-- For example, suppose we have something of type @(Int, [Int])@. The
+-- various contexts we might get are
+--
+-- > Context                  The thunk is..
+-- > ---------------------------------------------------------------------
+-- > ["(,)"]                  the pair itself
+-- > ["Int","(,)"]            the Int in the pair
+-- > ["List","(,)"]           the [Int] in the pair
+-- > ["Int","List","(,)"]     an Int in the [Int] in the pair
+--
+-- Note: prior to `ghc-9.6` a list was indicated by `[]`.
+data ThunkInfo = ThunkInfo {
       thunkContext :: Context
+    , thunkInfo    :: Maybe Info
     }
-  deriving (Show)
+  deriving Show
+
+-- | Construct `ThunkInfo` either from `Context` or information provided by
+-- `GHC` about `a` (see `whereFrom`).
+--
+mkThunkInfo :: Context -> a -> IO ThunkInfo
+#if MIN_VERSION_base(4,16,0)
+mkThunkInfo ctxt a = ThunkInfo ctxt . fmap fmt <$> whereFrom a
+  where
+    fmt :: InfoProv -> Info
+    fmt InfoProv { ipSrcFile, ipSrcSpan,
+                              ipLabel, ipTyDesc } =
+           ipLabel ++ " :: " ++ ipTyDesc
+        ++ " @ " ++ ipSrcFile ++ ":" ++ ipSrcSpan
+#else
+mkThunkInfo ctxt _ = return (ThunkInfo ctxt Nothing)
+#endif
+
 
 {-# NOINLINE unsafeNoThunks #-}
 -- | Call 'noThunks' in a pure context (relies on 'unsafePerformIO').
@@ -357,8 +395,9 @@ instance KnownSymbol name => NoThunks (InspectHeapNamed name a) where
 inspectHeap :: Context -> a -> IO (Maybe ThunkInfo)
 inspectHeap ctxt x = do
     containsThunks <- checkContainsThunks x
+    thunkInfo <- mkThunkInfo ("..." : ctxt) x
     return $ if containsThunks
-               then Just $ ThunkInfo { thunkContext = "..." : ctxt }
+               then Just thunkInfo
                else Nothing
 
 {-------------------------------------------------------------------------------
@@ -381,7 +420,7 @@ instance GWNoThunks a f => GWNoThunks a (D1 c f) where
 instance GWNoThunks a f => GWNoThunks a (C1 c f) where
   gwNoThunks a ctxt (M1 fp) = gwNoThunks a ctxt fp
 
-instance GWNoThunks a f => GWNoThunks a (S1 ('MetaSel ('Nothing) su ss ds) f) where
+instance GWNoThunks a f => GWNoThunks a (S1 ('MetaSel 'Nothing su ss ds) f) where
   gwNoThunks a ctxt (M1 fp) = gwNoThunks a ctxt fp
 
 instance (GWNoThunks a f, GWNoThunks a g) => GWNoThunks a (f :*: g) where
@@ -479,6 +518,45 @@ deriving via OnlyCheckWhnf Word16 instance NoThunks Word16
 deriving via OnlyCheckWhnf Word32 instance NoThunks Word32
 deriving via OnlyCheckWhnf Word64 instance NoThunks Word64
 
+{-------------------------------------------------------------------------------
+  Semigroups
+-------------------------------------------------------------------------------}
+
+deriving via a instance NoThunks a => NoThunks (Semigroup.Min a)
+deriving via a instance NoThunks a => NoThunks (Semigroup.Max a)
+deriving via a instance NoThunks a => NoThunks (Semigroup.First a)
+deriving via a instance NoThunks a => NoThunks (Semigroup.Last a)
+deriving via a instance NoThunks a => NoThunks (Semigroup.Dual a)
+deriving via Bool instance NoThunks Semigroup.All
+deriving via Bool instance NoThunks Semigroup.Any
+deriving via a instance NoThunks a => NoThunks (Semigroup.Sum a)
+deriving via a instance NoThunks a => NoThunks (Semigroup.Product a)
+deriving via a instance NoThunks a => NoThunks (Semigroup.WrappedMonoid a)
+instance (NoThunks a, NoThunks b) => NoThunks (Semigroup.Arg a b)
+
+{-------------------------------------------------------------------------------
+  Monoids
+-------------------------------------------------------------------------------}
+
+deriving via (Maybe a) instance NoThunks a => NoThunks (Monoid.First a)
+deriving via (Maybe a) instance NoThunks a => NoThunks (Monoid.Last a)
+deriving via (f a) instance NoThunks (f a) => NoThunks (Monoid.Alt f a)
+deriving via (f a) instance NoThunks (f a) => NoThunks (Monoid.Ap f a)
+
+{-------------------------------------------------------------------------------
+  Solo
+-------------------------------------------------------------------------------}
+
+#if MIN_VERSION_base(4,18,0)
+-- GHC-9.6 and newer
+instance NoThunks a => NoThunks (Solo a) where
+    wNoThunks ctx (MkSolo a) = wNoThunks ("Solo" : ctx) a
+#elif MIN_VERSION_base(4,16,0)
+-- GHC-9.2
+instance NoThunks a => NoThunks (Solo a) where
+    wNoThunks ctx (Solo a) = wNoThunks ("Solo" : ctx) a
+#endif
+
 {-------------------------------------------------------------------------------
   Mutable Vars
 -------------------------------------------------------------------------------}
